# -*- coding: utf-8 -*-
"""

Multi-Goal Motion Planner (MGMP)

@author: P.Vana & P.Cizek
"""
import sys
import os
import numpy as np

import math

import GridMap as gmap
import GridPlanner as gplanner

import matplotlib.pyplot as plt

from invoke_LKH import *

class MGMP_Solver:
    def __init__(self):
        pass

    # compute the shortest sequence based on the distance matrix (self.distances)
    def compute_TSP_sequence(self, start_idx = None, end_idx = None):
        n = len(self.distances)
        
        if start_idx != None and end_idx != None:
            # TODO - insert your code here
            print ("TODO")
            M = n * np.max(np.max(self.distances))
            for i in range(n):
                self.distances[i, start_idx] = M
                self.distances[i, end_idx] = M
            self.distances[end_idx, start_idx] = 0
        
        fname_tsp = "problem"
        user_comment = "a comment by the user"
        writeTSPLIBfile_FE(fname_tsp, self.distances,user_comment)
        run_LKHsolver_cmd(fname_tsp)
        sequence = read_LKHresult_cmd(fname_tsp)
        return sequence
    
    # compute the shortest tour based on the distance matrix (self.distances)
    def compute_TSP_tour(self, start_idx = None, end_idx = None):
        n = len(self.distances)
        sequence = self.compute_TSP_sequence(start_idx=start_idx, end_idx=end_idx)
        
        if start_idx != None and end_idx != None:
            # TODO - reconstruct only the open-loop
            print('TODO')
            exit(1)
        else:
            path = []
            for a in range(0,n):
                b = (a+1) % n
                a_idx = sequence[a]
                b_idx = sequence[b]
                actual_path = self.paths[(a_idx,b_idx)]
                path = path + actual_path
        
        return path
        
    
    def plan_tour_etsp(self, gridmap, goals, start_idx = None, end_idx = None):
        n = len(goals)
        self.distances = np.zeros((n,n))	
        self.paths = {}

        planner = gplanner.GridPlanner()

        # find path between each pair of goals (a,b)
        for a in range(0,n): 
            for b in range(0,n):
                g1 = goals[a]
                g2 = goals[b]
                if a != b:
                    # store distance
                    self.distances[a][b] = gridmap.dist_euclidean(g1, g2) 
                    # store path
                    self.paths[(a,b)] = [g1, g2]

        return self.compute_TSP_tour(start_idx, end_idx)
    
    def plan_tour_obstacles(self, gridmap, goals, start_idx = None, end_idx = None):	
        n = len(goals)
        self.distances = np.zeros((n,n))	
        self.paths = {}

        planner = gplanner.GridPlanner()

        # find path between each pair of goals (a,b)
        for a in range(0,n): 
            print('Shortest paths from city {0}'.format(a))
            for b in range(0,n):
                if a != b:
                    g1 = goals[a]
                    g2 = goals[b]
                
                    # find the shortest path
                    sh_path = planner.plan(gridmap, g1[:2], g2[:2])
                    # store distance
                    dist = 0
                    last = sh_path[0]
                    for act in sh_path:
                        dist += gridmap.dist_euclidean(last, act)
                        last = act 
                    self.distances[a][b] = dist
                    # store path
                    self.paths[(a,b)] = sh_path

        return self.compute_TSP_tour(start_idx, end_idx)
    
    def plan_tour_obstacles_Floyd_Warshall(self, gridmap, goals, start_idx = None, end_idx = None):	
        # TODO - reconstruct only the open-loop
        print('TODO')
        exit(1)
    
    def plan_tour_obstacles_neighborhood(self, gridmap, goals, neighbors, start_idx = None, end_idx = None):	
        n = len(goals)
        self.distances = np.zeros((n,n))	
        self.paths = {}

        planner = gplanner.GridPlanner()

        # compute sequence based on the non-collision Euclidean distances
        for a in range(0,n): 
            print('Shortest paths from city {0}'.format(a))
            for b in range(0,n):
                if a != b:
                    g1 = goals[a]
                    g2 = goals[b]
                
                    # find the shortest path
                    sh_path = planner.plan(gridmap, g1, g2, neigh = 'N8')
                    # store distance
                    dist = 0
                    last = sh_path[0]
                    for act in sh_path:
                        dist += gridmap.dist_euclidean(last, act)
                        last = act 
                    self.distances[a][b] = dist
                    # store path
                    self.paths[(a,b)] = sh_path

        sequence = self.compute_TSP_sequence(start_idx, end_idx)

        # TODO - this code select the first sample in the set!
        selected_samples = [0] * n

        path = []
        if start_idx != None and end_idx != None:
            # TODO - reconstruct only the open-loop
            print('TODO')
            exit(1)
        else:
            for a in range(0,n):
                b = (a+1) % n
                a_idx = sequence[a]
                b_idx = sequence[b]
                a_sample = neighbors[a_idx][selected_samples[a_idx]]
                b_sample = neighbors[b_idx][selected_samples[b_idx]]
                actual_path = planner.plan(gridmap, a_sample, b_sample, neigh = 'N8')
                path = path + actual_path

        return path
     
     
    
